// SPHERE_SELECT sphere_select.cpp
//
// Andrew Magis <amagis@cise.ufl.edu>
// Last change: Time-stamp: <31-Jul-2006 11:52:42 amagis>
//
// Version 1.0
// This software is open source, to be used in conjunction with the UCSF DOCK
// package.  If you have questions, comments, or suggestions, please contact
// me at the email address above.
//

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

//Global structs
struct Coordinates {
	int atom_number;
	int residue_number;
	double x;
	double y;
	double z;
} *coords = NULL;

struct Sphere {
	int this_atom;
	int pair_atom;
	double rtmin;
	double x, y, z;
	bool print;
} *spheres = NULL;

//Global variables
int total_spheres = 0;
int num_coords = 0;
double radius = -1.0;

//Function definitions
void Usage();
bool WithinRadius(Sphere *s, Coordinates *c);
void ReadSpheres(FILE *spherefile);
void ReadCoordinates(FILE *coordinatefile);
void WritePDB(FILE *outfile);

int main(int argc, char **argv)
{
	FILE *spherefile = NULL, *outfile = NULL, *coordinatefile = NULL;
	char spherefilename[200] = "",
		outputfilename[200] = "selected_spheres.sph",
		coordinatefilename[200] = "";
	int sphere_number = -1, printed_spheres = 0, atom_number = -1;
	char ch;
	bool write_pdb = false;
	
	while ((ch = getopt(argc, argv, "i:c:s:a:r:po:")) != -1) {

		switch(ch) {

		case 'i':
			strcpy(spherefilename, optarg);
			break;
		case 'c':
			strcpy(coordinatefilename, optarg);
			break;
		case 's':
			sphere_number = atoi(optarg);
			if (sphere_number < 0) {
				printf("Error! Invalid sphere number: %d\n", sphere_number);
				Usage();
			}
			break;
		case 'a':
			atom_number = atoi(optarg);
			if (atom_number  < 0) {
				printf("Error! Invalid atom number: %d\n", atom_number);
				Usage();
			}
			break;
		case 'r':
			radius = atof(optarg);
			break;
		case 'p':
			write_pdb = true;
			break;
		case 'o':
			strcpy(outputfilename, optarg);
			break;
		default:
			Usage();
			break;
		}
	}

	//Make sure radius is specifed on command line
	if (radius < 0.0) {
		fprintf(stderr, "Error! Selection radius must be specified\n\n");
		Usage();
	}

	//Make sure coordinate file exists if atom number is requested
	if ((atom_number != -1) && (strlen(coordinatefilename) == 0)) {
		fprintf(stderr, "Error! Coordinate file must be specified if atom name is specified\n\n");
	Usage();
	}

	//Make sure sphere number and atom_number are not both requested
	if ((atom_number != -1) && (sphere_number != -1)) {
		fprintf(stderr, "Error! Atom number and sphere number cannot both be specified\n\n");
	Usage();
	}
	
	//Try reading in spheres
	spherefile = fopen(spherefilename, "rt");
	if (spherefile == NULL) {
		fprintf(stderr, "Error! Cannot open sphere file '%s'\n",
				spherefilename);
		Usage();
	}
	ReadSpheres(spherefile);
	fclose(spherefile);
	printf("Read in %d spheres from '%s'\n", total_spheres, spherefilename);

	//Try reading in coordinates from coordinatefile
	if (sphere_number == -1) {

		coordinatefile = fopen(coordinatefilename, "rt");
		if (coordinatefile == NULL) {
			fprintf(stderr, "Error! Cannot open coordinate file '%s'\n",
					coordinatefilename);
			Usage();
		}
		ReadCoordinates(coordinatefile);
		fclose(coordinatefile);
		printf("Read in %d coordinates from '%s'\n", num_coords,
			   coordinatefilename);

		if (atom_number != -1) {
			
			//If particular atom is chosen, then select that one and
			//delete the rest
			int temp_coords = num_coords;
			num_coords = 0;
			for (int i = 0; i < temp_coords; i++) {

				if (coords[i].atom_number == atom_number) {
					Coordinates *c = new Coordinates;
					c[0].atom_number = coords[i].atom_number;
					c[0].x = coords[i].x;
					c[0].y = coords[i].y;
					c[0].z = coords[i].z;
					delete[] coords;
					coords = c;
					num_coords = 1;
					break;
				}
			}
			
			if (num_coords == 0) {
				fprintf(stderr, "Error! Atom %d not found in '%s'\n",
						atom_number, coordinatefilename);
				exit(1);
			}
			printf("Ignoring all coordinates except for atom %d\n",
				   atom_number);
		}
	}
	else {
		
		//If particular sphere is chosen, then copy that one
		for (int i = 0; i < total_spheres; i++) {
			if (sphere_number == spheres[i].this_atom) {
				coords = new Coordinates;
				coords[0].x = spheres[i].x;
				coords[0].y = spheres[i].y;
				coords[0].z = spheres[i].z;
				num_coords = 1;
				break;
			}
		}

		if (num_coords == 0) {
			fprintf(stderr, "Error! Sphere number %d not found in '%s'\n",
					sphere_number, spherefilename);
			exit(1);
		}
		printf("Using sphere %d in place of atom coordinates\n",
			   sphere_number);
	}
	
	//Parse through all spheres, compare distances between each sphere
	//and all possible coordinates, and write out the ones that pass
	for (int i = 0; i < total_spheres; i++) {
		spheres[i].print = false;
		for (int j = 0; j < num_coords; j++) {

			if (WithinRadius(&spheres[i], &coords[j])) {
				spheres[i].print = true;
				printed_spheres++;
				break;
			}
		}
	}
	
	//Open outfile for writing
	outfile = fopen(outputfilename, "wt");
	fprintf(outfile, "Spheres within %.3f angstroms of ligand/atom\n",
			radius);
	fprintf(outfile, "cluster%6d   number of spheres in cluster%6d\n",
			1, printed_spheres);

	//Print spheres to outfile
	for (int i = 0; i < total_spheres; i++) {
		if (spheres[i].print)
			fprintf(outfile, "%5d%10.5f%10.5f%10.5f%8.3f%5d%2d%3d\n",
					spheres[i].this_atom, spheres[i].x, spheres[i].y,
					spheres[i].z, spheres[i].rtmin, spheres[i].pair_atom,
					0, 0);
	}
	
	//Close output stream
	fclose(outfile);

	printf("%d spheres written to file '%s'\n", printed_spheres,
		   outputfilename);

	//If write_pdb is set, write out PDB file
	if (write_pdb) {
		outfile = fopen(strcat(outputfilename, ".pdb"), "wt");
		WritePDB(outfile);
		fclose(outfile);
		printf("PDB file written as '%s'\n", outputfilename);
	}
	
	if (coords != NULL) {
		if (num_coords == 1)
			delete coords;
		else delete[] coords;
	}
	if (spheres != NULL)
		delete[] spheres;
	
	return 0;
}

//Writes spheres out in PDB format
void WritePDB(FILE *outfile)
{
	for (int i = 0; i < total_spheres; i++) {
		if (spheres[i].print) {
			fprintf(outfile,"%-6s%5d %4s %3s  %4d    %8.3f%8.3f%8.3f\n",
					"ATOM", spheres[i].this_atom, "C", "SPH",
					spheres[i].this_atom, spheres[i].x, spheres[i].y,
					spheres[i].z);
			fprintf(outfile, "%-6s\n", "TER");
		}
	}
}

//Returns true if distance between sphere and coordinates is less
//than the radius
bool WithinRadius(Sphere *s, Coordinates *c)
{
	double distance;
	distance = sqrt( (c->x - s->x) * (c->x - s->x) +
					 (c->y - s->y) * (c->y - s->y) +
					 (c->z - s->z) * (c->z - s->z) );
	if (distance <= radius)
		return true;
	return false;
}

//Reads spheres in from the passed sphere file.  Does not
//read in spheres that are not part of a cluster (i.e. those
//in cluster 0)
void ReadSpheres(FILE *spherefile) {

	char line[500];
	int cluster = 0, num_spheres = 0, index = 0, items_read;

	while (!feof(spherefile)) {

		//Get a line from the sphere file
		fgets(line, 500, spherefile);

		//If this line is the beginning of a cluster
		if (strncmp(line, "cluster", 7) == 0) {
			items_read = sscanf(line, "%*s%d%*s%*s%*s%*s%*s%d", 
								&cluster, &num_spheres);

			//Sum spheres in all clusters
			if (cluster != 0) 
				total_spheres += num_spheres;
		}
	}					
	if (total_spheres == 0) {
		fprintf(stderr, "Error! Sphere file contains no spheres, or incorrect file format\n");
		exit(1);
	}

	//Allocate space for sphere array
	spheres = new Sphere[total_spheres];
	rewind(spherefile);

	while (!feof(spherefile)) {

		//Get a line from the sphere file
		fgets(line, 500, spherefile);

		//If this line is the beginning of a cluster
		if (strncmp(line, "cluster", 7) == 0) {
			items_read = sscanf(line, "%*s%d%*s%*s%*s%*s%*s%d", 
								&cluster, &num_spheres);
		}
		else {
			if (cluster != 0) {
				items_read = sscanf(line, "%d%lf%lf%lf%lf%d%*d%*d",
									&spheres[index].this_atom,
									&spheres[index].x,
									&spheres[index].y,
									&spheres[index].z,
									&spheres[index].rtmin,
									&spheres[index].pair_atom);
				index++;
			}
		}
	}
}

//Read coordinates in from coordinate file.  Supports MOL2 or PDB files
void ReadCoordinates(FILE *coordinatefile)
{
	bool is_mol2 = false;
	int items_read = 0;
	char line[500], atom_name[10];
	int atom_number;
	double x=0.0, y=0.0, z=0.0;
	int first=0, second=0, index=0;

	//Scan through file looking for @<TRIPOS>MOLECULE
	while (!feof(coordinatefile)) {
		items_read = fscanf(coordinatefile, "%17s", line);
		if (strncmp(line, "@<TRIPOS>MOLECULE", 17) == 0) {
			is_mol2 = true;
			break;
		}
	}

	//If is MOL2 format
	if (is_mol2) {
		printf("Coordinate file is MOL2 format\n");
		
		//Scan through file looking for @<TRIPOS>ATOM
		while (!feof(coordinatefile)) {
			items_read = fscanf(coordinatefile, "%13s", line);
			if (strncmp(line, "@<TRIPOS>ATOM", 13) == 0) {
				break;
			}
		}

		//Check to make sure not at end of file
		if (feof(coordinatefile)) {
			fprintf(stderr, "Error! MOL2 file has no atom coordinates, or '@<TRIIPOS>ATOM' marker not found\n");
			exit(1);
		}

		//Count number of coordinates in file
		first = ftell(coordinatefile);
		while (!feof(coordinatefile)) {

			fgets(line, 500, coordinatefile);
			items_read = sscanf(line, "%d%s%lf%lf%lf%*s",
								&atom_number, atom_name, &x, &y, &z);
			if (items_read == 5)
				num_coords++;
		}
		
		if (num_coords == 0) {
			fprintf(stderr, "Error! MOL2 file has no atom coordinates\n");
			exit(1);
		}
			
		second = ftell(coordinatefile);
		fseek(coordinatefile, first-second, SEEK_CUR);
		coords = new Coordinates[num_coords];

		//Scan file and collect atom coordinates		
		while (!feof(coordinatefile)) {

			fgets(line, 500, coordinatefile);
			items_read = sscanf(line, "%d%s%lf%lf%lf%*s",
								&atom_number, atom_name, &x, &y, &z);
			if (items_read == 5) {
				coords[index].atom_number = atom_number;
				coords[index].x = x;
				coords[index].y = y;
				coords[index].z = z;
				index++;
			}
		}
	}
	
	else {
		printf("Coordinate file is not MOL2 format: trying PDB\n");
		rewind(coordinatefile);

		//Count number of coordinates in file
		first = ftell(coordinatefile);
		while (!feof(coordinatefile)) {

			fgets(line, 500, coordinatefile);
			if ((strncmp(line, "ATOM", 4) == 0) ||
				(strncmp(line, "HETATM", 6) == 0)) 
				num_coords++;
		}

		if (num_coords == 0) {
			fprintf(stderr, "Error! PDB file has no atom coordinates, or file format not recognized\n");
			exit(1);
		}
		
		second = ftell(coordinatefile);
		fseek(coordinatefile, first-second, SEEK_CUR);
		coords = new Coordinates[num_coords];

		//Scan file and collect atom coordinates
		while (!feof(coordinatefile)) {

			fgets(line, 500, coordinatefile);
			if ((strncmp(line, "ATOM", 4) == 0) ||
				(strncmp(line, "HETATM", 6) == 0)) {

				items_read =
					sscanf(line, "%*6s%5d %4s%*c%*3s %*c%*4s%*c   %lf%lf%lf",
									&atom_number, atom_name, &x, &y, &z);
				
				if (items_read == 5) { 
					coords[index].atom_number = atom_number;
					coords[index].x = x;
					coords[index].y = y;
					coords[index].z = z;
					index++;
				}
				else {
					fprintf(stderr, "Error reading in line:\n%s\n", line);
				}
			}
		}
	}
}

//Prints program usage information
void Usage()
{
	printf("Usage: sphere_select -i inputfilename -r radius [-o outputfilename] [-c coordinatefilename] [-s sphere_number] [-a atom_number] [-p]\n\n");
	printf("\t-i\t Sphere input filename [required]\n");
	printf("\t-r\t Sphere selection radius [required]\n");
	printf("\t-o\t Output filename [default: selected_spheres.sph]\n");
	printf("\t-c\t Sphere selection coordinates [pdb or mol2 formats]\n");
	printf("\t-s\t Select spheres within radius around this sphere number\n");
	printf("\t  \t Ignores any coordinate file input\n");
	printf("\t-a\t Select spheres within radius around only this atom\n");
	printf("\t  \t from coordinate file\n");
	printf("\t-p\t Write out PDB file of spheres\n");
	exit(1);
}

